import os
import numpy as np
from ...annotations.base import BaseAnnotation
from ...annotations.manager import BaseAnnotationManager
from ..annotations import Label2D, Label3D
from ..base import EntityType


class DensityAnnotation(BaseAnnotation):
    """
    Base class for a Density Annotation

    All density annotations should derive from this base class and implement
    the 'draw' method. All derived classes will have access to the density
    instance (self.density) and all the annotation inputs and common params
    via self.interface.<property>

    An optional 'defaults' method can be provided to set default values
    to the annotation.

    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Auto-register any sub classes with the annotation manager
        DensityAnnotationManager.register(cls)

    def __init__(self, density):
        # Allow access to the density entity within the annotations
        self.density = density
        super().__init__()


class DensityAnnotationManager(BaseAnnotationManager):
    """
    Annotation Manager for Density Entity

    """

    _entity_type = EntityType.DENSITY
    _classes = {}  # Entity class specific annotation classes

    def __init__(self, entity):
        super().__init__(entity)
        self._interfaces = {}  # Entity instance specific annotation interfaces


class DensityInfo(DensityAnnotation):
    """
    Density Info Annotation

    Attributes
    ----------
    location: tuple[float, float]
        Normalized coordinates (0.0 - 1.0) of the postion in viewport / render

    show_filename: bool
        Whether or not to show the grid filename

    show_threshold: bool
        Whether or not to show the current threshold value

    show_origin: bool
        Whether or not to show the grid origin

    show_delta: bool
        Whether or not to show the grid delta

    show_shape: bool
        Whether or not to show the grid shape

    custom_text: str
        Any custom text to add at the end of the annotation

    """

    annotation_type = "density_info"

    location: tuple[float, float] = (0.025, 0.05)
    show_filename: bool = True
    show_threshold: bool = True
    show_origin: bool = True
    show_delta: bool = True
    show_shape: bool = True
    custom_text: str = ""

    def defaults(self) -> None:
        params = self.interface
        params.text_align = "left"

    def validate(self) -> bool:
        params = self.interface
        x, y = params.location
        if (not 0 <= x <= 1) or (not 0 <= y <= 1):
            raise ValueError("Normalized coordinates should lie between 0 and 1")
        return True

    def draw(self) -> None:
        params = self.interface
        grid = self.density.grid
        text = ""
        if params.show_filename:
            filename = os.path.basename(grid.metadata["filepath"])
            text = f"Filename: {filename}"
        if params.show_threshold:
            if self.density.node_group:
                iso_value = getattr(self.density.styles[0], "iso_value", None)
                if iso_value is None:
                    threshold = self.density.styles[0].threshold
                    text = text + f"|Threshold: {threshold:.2f}"
                else:
                    text = text + f"|ISO Value: {iso_value:.2f}"
        if params.show_origin:
            text = text + f"|Origin: {np.round(grid.origin, 2)}"
        if params.show_delta:
            text = text + f"|Delta: {np.round(grid.delta, 2)}"
        if params.show_shape:
            text = text + f"|Shape: {np.round(grid.grid.shape, 2)}"
        if params.custom_text != "":
            text = text + "|" + params.custom_text
        # Draw text at normalized coordinates wrt viewport / render
        self.draw_text_2d_norm(params.location, text)


class DensityGridAxes(DensityAnnotation):
    """
    Density Grid Axes Annotation

    Attributes
    ----------
    show_length: bool
        Whether or not to show the length of the grid axes

    units: str
        Units to use for length. Default: Å

    """

    annotation_type = "grid_axes"

    show_length: bool = True
    units: str = "Å"

    def defaults(self) -> None:
        params = self.interface
        params.text_depth = False
        params.line_arrow_size = 0.05
        params.text_color = (1.0, 1.0, 1.0, 0.5)
        params.line_color = (1.0, 1.0, 1.0, 0.5)

    def draw(self) -> None:
        params = self.interface
        grid = self.density.grid
        if grid.origin.size != 3:
            return
        origin = grid.origin.copy()
        if grid.metadata["center"]:
            origin = -np.array(grid.grid.shape) * 0.5 * grid.delta
        axes = ["X", "Y", "Z"]
        for i in range(3):
            length = grid.grid.shape[i] * grid.delta[i]
            mid_text = None
            if params.show_length:
                mid_text = f"{length:.2f} {params.units}"
            end = origin.copy()
            end[i] += length
            self.draw_line_3d(
                v1=origin, v2=end, mid_text=mid_text, v2_text=axes[i], v2_arrow=True
            )


class DensityGridAxes3D(DensityAnnotation):
    """
    Density Grid Axes 3D Annotation

    Attributes
    ----------
    show_length: bool
        Whether or not to show the length of the grid axes

    units: str
        Units to use for length. Default: Å

    """

    annotation_type = "grid_axes_3d"

    show_length: bool = True
    units: str = "Å"

    def defaults(self) -> None:
        params = self.interface
        params.text_depth = False
        params.mesh_color = (1.0, 1.0, 1.0, 0.5)

    def draw(self) -> None:
        params = self.interface
        grid = self.density.grid
        if grid.origin.size != 3:
            return
        origin = grid.origin.copy()
        if grid.metadata["center"]:
            origin = -np.array(grid.grid.shape) * 0.5 * grid.delta
        axes = ["X", "Y", "Z"]
        colors = [(1, 0, 0, 1), (0, 1, 0, 1), (0, 0, 1, 1)]
        for i in range(3):
            length = grid.grid.shape[i] * grid.delta[i]
            mid_text = None
            if params.show_length:
                mid_text = f"{length:.2f} {params.units}"
            # cone as arrow end - 10% of axis length
            cone_height = 0.1 * length
            cylinder_end = origin.copy()
            cylinder_end[i] += length - cone_height
            axis = cylinder_end - origin
            radius = 0.01 * length
            # draw axis as a cylinder
            overrides = {"mesh_color": colors[i]}
            self.draw_cylinder(
                origin, radius, length - cone_height, axis, overrides=overrides
            )
            # draw arrow end as a cone
            self.draw_cone(
                cylinder_end, radius * 3, cone_height, axis, overrides=overrides
            )
            mid = (origin + cylinder_end) / 2
            # draw axis length text
            if mid_text is not None:
                self.draw_text_3d(mid, mid_text)
            cylinder_end[i] += cone_height
            # draw axis name text
            self.draw_text_3d(cylinder_end, axes[i])


class Label2D(DensityAnnotation, Label2D):
    """Common Label2D Annotation for all entities"""


class Label3D(DensityAnnotation, Label3D):
    """Common Label3D Annotation for all entities"""



import pytest
import molecularnodes as mn
from .constants import data_dir


@pytest.fixture
def filepath():
    return data_dir / "1f2n.bcif"


def test_bcif_init(filepath):
    bcif = mn.Molecule.load(filepath)
    assert bcif._reader.file_path == filepath


def test_bcif_assemblies(filepath):
    bcif = mn.Molecule.load(filepath)
    assemblies = bcif.assemblies()
    assert assemblies is not None


def test_bcif_entity_ids(filepath):
    mol = mn.Molecule.load(filepath)
    entity_ids = list(mol.object["entity_ids"])
    assert entity_ids is not None
    assert entity_ids == ["CAPSID PROTEIN", "CALCIUM ION", "water"]



"""
Importing molecular dynamics trajectories and associated files.
"""

__name__ = "MolecularNodes.trajectory"
__author__ = "Brady Johnston"

from pathlib import Path
import MDAnalysis as mda
from . import StreamingTrajectory, imd, oxdna
from .base import Trajectory


def load(
    top: str | Path,
    traj: str | Path,
    name: str = "NewTrajectory",
    style: str | None = "spheres",
    selection: str | None = None,
):
    """Load a trajectory with automatic format detection.

    Detects and loads trajectories in various formats including standard
    MD trajectories and real-time IMD streaming connections.

    Parameters
    ----------
    top : str | Path
        Path to topology file
    traj : str | Path
        Path to trajectory file or IMD URL (e.g., 'imd://localhost:3000')
    name : str, optional
        Name for the trajectory object, by default "NewTrajectory"
    style : str | None, optional
        Visual style to apply, by default "spheres"
    selection : str | None, optional
        Atom selection string, by default None (all atoms)

    Returns
    -------
    Trajectory
        Loaded trajectory object (may be StreamingTrajectory for IMD)
    """
    if imd.is_imd_url(traj):
        trajectory = StreamingTrajectory.load(top, traj, name=name)
    else:
        universe = mda.Universe(top, traj)
        trajectory = Trajectory(universe=universe, name=name)

    trajectory.add_style(style=style, selection=selection)

    return trajectory


def load_oxdna(top, traj, name="oxDNA", style="oxdna", world_scale=0.01):
    """
    Load an oxDNA trajectory.

    Parameters
    ----------
    top : str
        Path to topology file
    traj : str
        Path to trajectory file
    name : str, optional
        Name for the created object, by default "oxDNA"
    style : str, optional
        Style of representation, by default "oxdna"
    world_scale : float, optional
        Scaling factor for world coordinates, by default 0.01

    Returns
    -------
    OXDNA
        The created trajectory object
    """
    univ = mda.Universe(
        top, traj, topology_format=oxdna.OXDNAParser, format=oxdna.OXDNAReader
    )
    traj = oxdna.OXDNA(univ, name=name, world_scale=world_scale).add_style(style=style)
    return traj



import itertools
import bpy
import numpy as np
import pytest
from databpy import ObjectTracker
import molecularnodes as mn
from .constants import codes, data_dir
from .utils import NumpySnapshotExtension


@pytest.mark.parametrize("code", codes)
def test_op_fetch(snapshot_custom: NumpySnapshotExtension, code):
    scene = bpy.context.scene
    style = "ribbon"
    format = "cif"

    with ObjectTracker() as o:
        bpy.ops.mn.import_fetch(code=code, file_format=format, style=style)
        mol1 = scene.MNSession.match(o.latest())

    with ObjectTracker() as o:
        bpy.ops.mn.import_fetch(
            code=f"pdb_{code.rjust(8, '0')}", file_format=format, style=style
        )
        mol2 = scene.MNSession.match(o.latest())

    mol3 = mn.Molecule.fetch(code, format=format, cache=data_dir)
    mol3.add_style(style=style)

    for test1, test2 in itertools.combinations([mol1, mol2, mol3], 2):
        np.testing.assert_allclose(test1.position, test2.position)


# TODO: Fix intermittent test failures
# def test_op_fetch_alphafold(tmpdir):
#     scene = bpy.context.scene
#     style = "ribbon"
#     code = "K4PA18"

#     with ObjectTracker() as o:
#         bpy.ops.mn.import_fetch(
#             code=code,
#             style=style,
#             cache_dir=str(tmpdir),
#             database="alphafold",
#         )
#         mol = scene.MNSession.match(o.latest())
#         assert mol._entity_type == mn.entities.base.EntityType.MOLECULE
#         assert mol.object.mn.entity_type == mol._entity_type.value

#     assert mol.name == code


@pytest.mark.parametrize("code", codes)
@pytest.mark.parametrize("file_format", ["bcif", "cif", "pdb"])
def test_op_local(snapshot_custom, code, file_format):
    session = bpy.context.scene.MNSession
    path = mn.download.StructureDownloader(cache=data_dir).download(
        code=code, format=file_format
    )

    with ObjectTracker() as o:
        bpy.ops.mn.import_local(filepath=str(path), node_setup=False)
        mol = session.match(o.latest())
        assert mol._entity_type == mn.entities.base.EntityType.MOLECULE
        assert mol.object.mn.entity_type == mol._entity_type.value

    with ObjectTracker() as o:
        bpy.ops.mn.import_local(filepath=str(path), centre=True, centre_type="centroid")
        mol_cent = session.match(o.latest())

    assert snapshot_custom == mol.position
    assert snapshot_custom == mol_cent.position
    assert not np.allclose(mol.position, mol_cent.position)


def test_op_api_mda(snapshot_custom: NumpySnapshotExtension):
    bpy.context.scene.frame_set(0)

    topo = str(data_dir / "md_ppr/box.gro")
    traj = str(data_dir / "md_ppr/first_5_frames.xtc")
    name = "AnotherNewTrajectory"

    with ObjectTracker() as o:
        bpy.ops.mn.import_trajectory(
            topology=topo, trajectory=traj, name=name, style="ribbon"
        )
        obj_1 = o.latest()

    traj_op = bpy.context.scene.MNSession.match(obj_1)
    assert traj_op.name == name
    assert traj_op._mn_entity_type == mn.entities.base.EntityType.MD.value

    traj_func = mn.entities.trajectory.load(topo, traj, name="test", style="ribbon")

    bpy.context.scene.frame_set(2)
    assert np.allclose(traj_func.position, traj_op.position)
    pos_2 = traj_func.position.copy()
    bpy.context.scene.frame_set(4)
    traj_op.set_frame(4)
    traj_func.set_frame(4)

    assert not np.allclose(pos_2, traj_op.position)
    assert not np.allclose(pos_2, traj_func.position)


@pytest.mark.skipif(
    bpy.app.version_string.startswith("4.2"),
    reason="Test fails in 4.2 but succeeeds otherwise",
)
def test_op_residues_selection_custom():
    topo = str(data_dir / "md_ppr/box.gro")
    traj = str(data_dir / "md_ppr/first_5_frames.xtc")

    with ObjectTracker():
        bpy.ops.mn.import_trajectory(
            topology=topo, trajectory=traj, name="NewTrajectory", style="ribbon"
        )
    area = bpy.context.screen.areas[-1]
    area.ui_type = "GeometryNodeTree"
    with bpy.context.temp_override(area=area):
        bpy.ops.mn.residues_selection_custom("EXEC_DEFAULT")



import platform
import numpy as np
import pytest
import molecularnodes as mn
from .utils import NumpySnapshotExtension


@pytest.mark.skipif(
    platform.system() != "Darwin" or platform.machine() != "arm64",
    reason="Test only runs on macOS ARM",
)
def test_random_rgb(snapshot_custom):
    n = 100
    colors = np.array(list(map(mn.color.random_rgb, range(n))))
    assert snapshot_custom == colors


def test_colos_from_atomic_numbers(snapshot_custom: NumpySnapshotExtension):
    length = len(mn.color.iupac_colors_rgb)
    colors = mn.color.colors_from_elements(np.array(list(range(length))))
    assert snapshot_custom == colors



STYLE_ITEMS = (
    ("spheres", "Spheres", "Space-filling atoms style."),
    ("cartoon", "Cartoon", "Secondary structure cartoons"),
    ("surface", "Surface", "Solvent-accsible surface."),
    ("ribbon", "Ribbon", "Continuous backbone ribbon."),
    ("sticks", "Sticks", "Sticks for each bond."),
    ("ball_and_stick", "Ball and Stick", "Spheres for atoms, sticks for bonds"),
    ("preset_1", "Preset 1", "A pre-made combination of different styles"),
    ("preset_2", "Preset 2", "A pre-made combination of different styles"),
    ("preset_3", "Preset 3", "A pre-made combination of different styles"),
    ("preset_4", "Preset 4", "A pre-made combination of different styles"),
)



import warnings
from abc import ABC
import bpy
from .render import enable_optimal_gpu


class RenderEngine(ABC):
    _name = "RenderEngineName"

    @property
    def name(self) -> str:
        return self._name

    def _enable_engine(self):
        bpy.context.scene.render.engine = self.name  # type: ignore


class EEVEE(RenderEngine):
    def __init__(self, samples: int = 64, raytracing: bool = True):
        # TODO: remove this check when Blender 5.0 is the minimum version
        if bpy.app.version >= (5, 0, 0):
            self._name = "BLENDER_EEVEE"
        else:
            self._name = "BLENDER_EEVEE_NEXT"
        self.samples = samples
        self.raytracing = raytracing

    @property
    def engine(self):
        return bpy.context.scene.eevee

    @property
    def samples(self) -> int:
        return self.engine.taa_render_samples

    @samples.setter
    def samples(self, value: int) -> None:
        self.engine.taa_render_samples = value

    @property
    def raytracing(self) -> bool:
        return self.engine.use_raytracing

    @raytracing.setter
    def raytracing(self, value: bool) -> None:
        self.engine.use_raytracing = value


class Cycles(RenderEngine):
    def __init__(
        self,
        samples: int = 256,
        device: str = "GPU",
        denoise: bool = True,
        denoise_gpu: bool = True,
    ):
        self._name = "CYCLES"
        self.samples = samples
        self.device = device
        self.denoise = denoise
        self.denoise_gpu = denoise_gpu
        self._enable_engine()

    @property
    def engine(self):
        return bpy.context.scene.cycles

    @property
    def samples(self) -> int:
        return self.engine.samples

    @samples.setter
    def samples(self, value: int) -> None:
        self.engine.samples = value

    @property
    def device(self) -> str:
        return self.engine.device

    @device.setter
    def device(self, value: str):
        value = value.upper()
        self.engine.device = value
        if value == "GPU":
            try:
                enable_optimal_gpu()
            except TypeError:
                warnings.warn(
                    "Failed to enable GPU, defaulting back to CPU render device"
                )
                self.engine.device = "CPU"

    @property
    def denoise(self) -> bool:
        return self.engine.use_denoising

    @denoise.setter
    def denoise(self, value: bool) -> None:
        self.engine.use_denoising = value

    @property
    def denoise_gpu(self) -> bool:
        return self.engine.denoising_use_gpu

    @denoise_gpu.setter
    def denoise_gpu(self, value: bool) -> None:
        self.engine.denoising_use_gpu = value



import bpy
import numpy as np
from bpy.app.handlers import persistent
from PIL import Image
from .annotations.utils import render_annotations
from .scene.compositor import annotations_image


# this update function requires a self and context input, as funcitons with these inputs
# have ot be passed to the `update` arguments of UI properties. When the UI is updated,
# the function is called with the UI element being `self` and the current context being
# passed into the function
def _update_entities(self, context: bpy.types.Context) -> None:
    """
    Function for being called at various points in the updating of the UI, to ensure
    positions and selections of the trajectories are udpated with the new inputs
    """
    update_entities(context.scene)


# this is the 'perisisent' function which can be appended onto the
# `bpy.app.handlers.frame_change_*` functions. Either before or after the frame changes
# this function will then be called - ensuring all of the trajectories are up to date. We
# use the `frame_change_pre` handler as we want the frame to change first, then we update
# the universe based on the current frame value
@persistent
def update_entities(scene):
    "Call the `set_frame()` method of all entities in the current session"
    session = scene.MNSession
    session.prune()

    for entity in session.entities.values():
        if entity.update_with_scene:
            frame_to_set = scene.frame_current
        else:
            frame_to_set = entity.frame

        # do the entity setting, if the method isn't implemented, just pass
        try:
            entity.set_frame(frame_to_set)
        except NotImplementedError:
            pass


@persistent
def render_pre_handler(scene: bpy.types.Scene) -> None:
    """
    Blender's on render (before) handler
    Any changes needed before the rendering of a frame need to go in here

    """
    # frame_change_pre is supposed to be called before render_pre
    # however, that doesn't seem to be the case
    # update entities here to ensure annotations match entity frame
    update_entities(scene)
    # Render annotations to an image
    bpy.context.view_layer.update()
    render_scale = scene.render.resolution_percentage / 100
    width = int(scene.render.resolution_x * render_scale)
    height = int(scene.render.resolution_y * render_scale)
    image_scale = 2  # To workaround anti-aliasing issues with lines
    image = Image.new(
        "RGBA",
        (width * image_scale, height * image_scale),
        (0, 0, 0, 0),
    )
    # render annotations of all entities
    render_annotations(scene, render_scale, image, image_scale)
    if image_scale != 1:
        # scale down to actual render size
        image = image.resize((width, height))
    # create blender image
    if annotations_image not in bpy.data.images:
        bpy.data.images.new(annotations_image, width, height)
    bpy_image = bpy.data.images[annotations_image]
    bpy_image.scale(width, height)
    # update from PIL image
    bpy_image.pixels[:] = (np.flipud(np.array(image).astype(float)) / 255.0).ravel()



from math import degrees, radians
from typing import Literal, Sequence, get_args
import bpy

Viewpoints = Literal["default", "front", "back", "top", "bottom", "left", "right"]

_viewpoint_rotation_eulers = {
    # "default" is the camera rotation as per the template
    "default": (radians(70.402), radians(0), radians(0)),
    "front": (radians(90), radians(0), radians(0)),
    "back": (radians(90), radians(0), radians(-180)),
    "top": (radians(0), radians(0), radians(0)),
    "bottom": (radians(-180), radians(0), radians(0)),
    "left": (radians(-270), radians(0), radians(-90)),
    "right": (radians(-270), radians(0), radians(-270)),
}


class Camera:
    """
    A class to handle camera settings in Blender.

    """

    def __init__(self):
        # set defaults that match viewport virtual camera
        self.lens = 50
        self.clip_start = 0.01
        self.clip_end = 1000

    @property
    def camera(self) -> bpy.types.Camera:
        """Get Camera object"""
        return bpy.context.scene.camera

    @property
    def camera_data(self) -> bpy.types.Camera:
        """Get Camera data"""
        return self.camera.data

    @property
    def lens(self) -> float:
        """Get Camera focal length"""
        return self.camera_data.lens

    @lens.setter
    def lens(self, value) -> None:
        """Set Camera focal length"""
        self.camera_data.lens = value

    @property
    def clip_start(self) -> float:
        """Get Camera near clipping distance"""
        return self.camera_data.clip_start

    @clip_start.setter
    def clip_start(self, value) -> None:
        """Set Camera near clipping distance"""
        self.camera_data.clip_start = value

    @property
    def clip_end(self) -> float:
        """Get Camera far clipping distance"""
        return self.camera_data.clip_end

    @clip_end.setter
    def clip_end(self, value) -> None:
        """Set Camera far clipping distance"""
        self.camera_data.clip_end = value

    @property
    def rotation(self) -> tuple[float, float, float]:
        """Get Camera rotation in degrees (XYZ)"""
        return tuple(degrees(angle) for angle in self.camera.rotation_euler)

    @rotation.setter
    def rotation(self, angles: tuple[float, float, float]) -> None:
        """Set Camera rotation in degrees (XYZ)"""
        self.camera.rotation_euler = tuple(radians(angle) for angle in angles)

    def set_viewpoint(self, viewpoint: Viewpoints | Sequence[float]) -> None:
        """
        Set viewpoint to a preset or a custom Euler rotation.

        Parameters
        ----------
        viewpoint : Viewpoints | Sequence[float]
            Either a named viewpoint string (e.g. "front", "top") or a tuple/list of three Euler angles in radians.
        """
        if isinstance(viewpoint, str):
            if viewpoint not in get_args(Viewpoints):
                raise ValueError(f"{viewpoint} is not one of {get_args(Viewpoints)}")
            self.camera.rotation_euler = _viewpoint_rotation_eulers[viewpoint]
        else:
            self.camera.rotation_euler = viewpoint



from pathlib import Path
from typing import List, Optional, Union
import bpy
import databpy
import mrcfile
import numpy as np
import starfile
from databpy import AttributeTypes, BlenderObject
from pandas import CategoricalDtype, DataFrame
from PIL import Image
from scipy.spatial.transform import Rotation
from ... import blender as bl
from ...nodes import nodes
from .base import Ensemble, EntityType


class EnsembleDataFrame:
    def __init__(self, data: DataFrame) -> None:
        self.data = data
        self._coord_columns: List[str] = ["x", "y", "z"]
        self._rot_columns: List[str] = ["Rot", "Tilt", "Psi"]
        self._shift_column_names: List[str] = [
            "OriginXAngst",
            "OriginYAngst",
            "OriginZAngst",
        ]

    @property
    def coordinates(self) -> np.ndarray:
        coord = self.data[self._coord_columns].to_numpy()

        try:
            shift = self.data[self._shift_column_names].to_numpy()
            coord -= shift
        except KeyError:
            pass

        return coord

    @property
    def scale(self) -> np.ndarray:
        arr = np.zeros((len(self.data), 1), dtype=np.float32)
        arr[:] = 1.0
        return arr

    @property
    def coordinates_scaled(self) -> np.ndarray:
        return self.coordinates * self.scale

    def rotation_as_quaternion(self) -> np.ndarray:
        rot_tilt_psi_cols = self.data[self._rot_columns].to_numpy()

        quaternions = np.array(
            [
                Rotation.from_euler("ZYZ", row, degrees=True)
                .inv()
                .as_quat(scalar_first=True)
                for row in rot_tilt_psi_cols
            ]
        )
        return quaternions

    def image_id_values(self) -> np.ndarray:
        for name in [
            "rlnImageName",
            "rlnMicrographName",
            "rlnTomoName",
            "cisTEMOriginalImageFilename",
        ]:
            try:
                return self.data[name].cat.codes.to_numpy()
            except KeyError:
                pass

        return np.zeros(len(self.data), dtype=int)

    def store_data_on_object(self, obj: bpy.types.Object) -> None:
        bob = BlenderObject(obj)
        bob.store_named_attribute(
            self.rotation_as_quaternion(),
            name="rotation",
            atype=AttributeTypes.QUATERNION,
        )

        bob.store_named_attribute(
            self.image_id_values(),
            name="image_id",
            atype=AttributeTypes.INT,
        )

        for col in self.data.columns:
            if isinstance(self.data[col].dtype, CategoricalDtype):
                bob.object[f"{col}_categories"] = list(self.data[col].cat.categories)
                data = self.data[col].cat.codes.to_numpy()
                bob.store_named_attribute(data, name=col, atype=AttributeTypes.INT)
            else:
                bob.store_named_attribute(self.data[col].to_numpy(), name=col)


class RelionDataFrame(EnsembleDataFrame):
    def __init__(self, data: DataFrame) -> None:
        super().__init__(data)
        self.type: str = "relion"
        self._coord_columns = ["rlnCoordinateX", "rlnCoordinateY", "rlnCoordinateZ"]
        self._rot_columns = ["rlnAngleRot", "rlnAngleTilt", "rlnAnglePsi"]
        self._shift_column_names = [
            "rlnOriginXAngst",
            "rlnOriginYAngst",
            "rlnOriginZAngst",
        ]

    @property
    def scale(self) -> np.ndarray:
        if "rlnImagePixelSize" not in self.data:
            return super().scale

        return self.data["rlnImagePixelSize"].to_numpy().reshape((-1, 1))


class CistemDataFrame(EnsembleDataFrame):
    def __init__(self, data: DataFrame) -> None:
        super().__init__(data)
        self._adjust_defocus()
        self.type: str = "cistem"
        self._coord_columns = [
            "cisTEMOriginalXPosition",
            "cisTEMOriginalYPosition",
            "cisTEMZFromDefocus",
        ]
        self._rot_columns = ["cisTEMAnglePhi", "cisTEMAngleTheta", "cisTEMAnglePsi"]
        self._shift_column_names = [
            "origin_x",
            "origin_y",
            "origin_z",
        ]

    def _adjust_defocus(self) -> None:
        self.data["cisTEMZFromDefocus"] = (
            self.data["cisTEMDefocus1"] + self.data["cisTEMDefocus2"]
        ) / 2
        self.data["cisTEMZFromDefocus"] = (
            self.data["cisTEMZFromDefocus"] - self.data["cisTEMZFromDefocus"].median()
        )


class StarFile(Ensemble):
    data_reader: Optional[DataFrame]
    data_frame: Optional[Union[RelionDataFrame, CistemDataFrame]]

    def __init__(self, file_path: Union[str, Path]) -> None:
        super().__init__(file_path)
        self.type: str = "starfile"
        self.current_image: int = -1
        self._entity_type = EntityType.ENSEMBLE_STAR

    @classmethod
    def from_starfile(cls, file_path: Union[str, Path]) -> "StarFile":
        self = cls(file_path)
        self.data_reader = self._read()
        self.data_frame = self._assign_df()
        return self

    @classmethod
    def from_blender_object(cls, blender_object: bpy.types.Object) -> "StarFile":
        self = cls(blender_object["starfile_path"])
        self.object = blender_object
        self.data_reader = self._read()
        return self

    @property
    def star_node(self) -> bpy.types.Node:
        return nodes.get_star_node(self.object)

    @property
    def micrograph_material(self) -> bpy.types.Material:
        return nodes.micrograph_material()

    def _read(self) -> DataFrame:
        star_dict: dict = starfile.read(self.file_path, always_dict=True)  # type: ignore
        star: DataFrame = list(star_dict.values())[0]

        if not isinstance(star, DataFrame):
            raise ValueError("Problem opening starfile as dataframe")

        for col in star.columns:
            if star[col].dtype == "object":
                star[col] = star[col].astype("category")
        return star

    @property
    def n_images(self) -> int:
        if isinstance(self.data_reader, dict):
            return len(self.data_reader)
        return 1

    def _is_relion(self) -> bool:
        return (
            isinstance(self.data_reader, dict)
            and "particles" in self.data_reader
            and "optics" in self.data_reader
        ) or ("rlnAnglePsi" in self.data_reader)  # type: ignore

    def _is_cistem(self) -> bool:
        return "cisTEMAnglePsi" in self.data_reader  # type: ignore

    def _assign_df(self) -> Union[RelionDataFrame, CistemDataFrame]:
        if self.data_reader is None:
            raise ValueError("Data not loaded. Call from_starfile() first.")
        if self._is_relion():
            return RelionDataFrame(self.data_reader)
        elif self._is_cistem():
            return CistemDataFrame(self.data_reader)
        else:
            raise ValueError(
                "File is not a valid RELION>=3.1 or cisTEM STAR file, other formats are not currently supported."
            )

    def _convert_mrc_to_tiff(self) -> Path:
        if self.object is None:
            raise ValueError("Object not set. Call from_blender_object() first.")

        if self._is_relion():
            micrograph_path = self.object["rlnMicrographName_categories"][
                self.star_node.inputs["Image"].default_value - 1  # type: ignore
            ]
        elif self._is_cistem():
            micrograph_path = self.object["cisTEMOriginalImageFilename_categories"][
                self.star_node.inputs["Image"].default_value - 1  # type: ignore
            ].strip("'")
        else:
            raise ValueError("File is not a valid RELION>=3.1 or cisTEM STAR file")

        micrograph_path = Path(micrograph_path)
        if not micrograph_path.exists():
            pot_micrograph_path = Path(self.file_path).parent / micrograph_path
            if not pot_micrograph_path.exists():
                if self._is_relion():
                    pot_micrograph_path = (
                        Path(self.file_path).parent.parent.parent / micrograph_path
                    )
                    if not pot_micrograph_path.exists():
                        raise FileNotFoundError(
                            f"Micrograph file {micrograph_path} not found"
                        )
                else:
                    raise FileNotFoundError(
                        f"Micrograph file {micrograph_path} not found"
                    )
            micrograph_path = pot_micrograph_path

        tiff_path = Path(micrograph_path).with_suffix(".tiff")
        if not tiff_path.exists():
            with mrcfile.open(micrograph_path) as mrc:
                if mrc.data is None:
                    raise ValueError(f"Micrograph file {micrograph_path} is empty")
                micrograph_data = mrc.data.copy()

            if micrograph_data.ndim == 3:
                micrograph_data = np.sum(micrograph_data, axis=0)
            micrograph_data = (micrograph_data - micrograph_data.min()) / (
                micrograph_data.max() - micrograph_data.min()
            )

            if micrograph_data.dtype != np.float32:
                micrograph_data = micrograph_data.astype(np.float32)

            Image.fromarray(micrograph_data[::-1, :]).save(tiff_path)
        return tiff_path

    def create_object(
        self,
        name: str = "StarFileObject",
        node_setup: bool = True,
        world_scale: float = 0.01,
        fraction: float = 1.0,
        simplify: bool = True,
    ) -> bpy.types.Object:
        if self.data_frame is None:
            raise ValueError("DataFrame not assigned. Call from_starfile() first.")

        self.object = databpy.create_object(
            self.data_frame.coordinates_scaled * world_scale,
            collection=bl.coll.mn(),
            name=name,
        )
        self.object.mn.entity_type = self._entity_type.value
        self.data_frame.store_data_on_object(self.object)

        if node_setup:
            nodes.create_starting_nodes_starfile(self.object)

        self.object["starfile_path"] = str(self.file_path)
        return self.object



import bpy
import MDAnalysis as mda
import pytest
import molecularnodes as mn
from .constants import data_dir


@pytest.fixture
def canvas():
    return mn.scene.base.Canvas()


@pytest.fixture()
def universe():
    topo = data_dir / "md_ppr/box.gro"
    traj = data_dir / "md_ppr/first_5_frames.xtc"
    return mda.Universe(topo, traj)


def test_resolution(canvas):
    # Test getting and setting resolution
    canvas.resolution = (1920, 1080)
    assert canvas.resolution == (1920, 1080)


def test_load_blend():
    # Test loading a .blend file
    file = data_dir / "blendfiles/suzanne.blend"
    assert not bpy.data.objects.get("Suzanne")
    canvas = mn.Canvas(template=file)
    assert bpy.data.objects.get("Suzanne")
    canvas.scene_reset(None)
    assert not bpy.data.objects.get("Suzanne")
    assert bpy.data.objects.get("Cube")
    canvas.load(file)
    assert bpy.data.objects.get("Suzanne")
    with pytest.raises(ValueError):
        mn.Canvas(template="x")


def test_animation_settings(canvas):
    # Test FPS
    canvas.fps = 30
    assert canvas.fps == 30

    # Test frame range
    canvas.frame_start = 1
    canvas.frame_end = 250
    assert canvas.frame_start == 1
    assert canvas.frame_end == 250


def test_frame_object(canvas):
    # Create test object
    bpy.ops.mesh.primitive_cube_add()
    test_obj = bpy.context.active_object

    # Test framing
    canvas.frame_object(test_obj)

    # Cleanup
    bpy.data.objects.remove(test_obj, do_unlink=True)


def test_frame_view(canvas, universe):
    t1 = mn.Trajectory(universe)

    # save views for later framing
    # view of resid 1
    v1 = t1.get_view(selection="resid 1")
    # view of resid 1 at trajectory frame 3
    v13 = t1.get_view(selection="resid 1", frame=3)
    # view of resid 2
    v2 = t1.get_view(selection="resid 2")
    # view of whole trajectory
    v0 = t1.get_view()

    camera = bpy.context.scene.camera
    initial_location = camera.location.copy()

    # frame v1 (resid 1 of trajectory)
    canvas.frame_view(v1)
    l1 = camera.location.copy()
    assert l1 != initial_location
    # frame v13 (resid 1 of trajectory at frame 3)
    canvas.frame_view(v13)
    l13 = camera.location.copy()
    assert l13 != l1
    # frame v2 (resid 2 of trajectory)
    canvas.frame_view(v2)
    l2 = camera.location.copy()
    assert l2 != l1 and l2 != l13
    # frame whole trajectory
    canvas.frame_view(v0)
    l0 = camera.location.copy()
    assert l0 != initial_location

    # test different viewpoints
    canvas.frame_view(v1, viewpoint="front")
    r1f = camera.rotation_euler.copy()
    canvas.frame_view(v1, viewpoint="back")
    r1b = camera.rotation_euler.copy()
    assert r1f != r1b

    # test framing multiple views
    # frame v1 and v2 (resid 1 and resid 2 of trajectory)
    canvas.frame_view(v1 + v2)
    l12 = camera.location.copy()
    assert l12 != l1 and l12 != l2

    # test camera lens (zoom in and out)
    # location should not change until re-framed
    canvas.camera.lens = 35
    l3 = camera.location.copy()
    assert l3 == l12
    canvas.camera.lens = 85
    l4 = camera.location.copy()
    assert l4 == l3 and l4 == l12


def test_compositor_setup(canvas):
    if bpy.app.version >= (5, 0, 0):
        node_tree = canvas.scene.compositing_node_group
        output_node_name = "Group Output"
    else:
        assert canvas.scene.use_nodes
        node_tree = canvas.scene.node_tree
        output_node_name = "Composite"
    assert node_tree is not None
    mn_compositor_node_name = mn.scene.compositor.mn_compositor_node_name
    assert mn_compositor_node_name in node_tree.nodes
    mn_compositor_node = node_tree.nodes[mn_compositor_node_name]
    assert mn_compositor_node.inputs["Image"].is_linked
    assert mn_compositor_node.outputs["Image"].is_linked
    assert mn_compositor_node.outputs["Image"].links[0].to_node.name == output_node_name



from .base import Ensemble
from .cellpack import CellPack
from .io import load_cellpack, load_starfile
from .star import StarFile

__all__ = [
    "Ensemble",
    "CellPack",
    "load_cellpack",
    "load_starfile",
    "StarFile",
]



import typing
from collections import Counter, deque
import bpy
from mathutils import Vector


def contains_geo_socket(sockets: bpy.types.NodeInputs | bpy.types.NodeOutputs) -> bool:
    """
    Check if any socket in the collection is a geometry socket

    Parameters
    ----------
    sockets : bpy.types.NodeInputs | bpy.types.NodeOutputs
        Collection of node sockets to check

    Returns
    -------
    bool
        True if any socket is a geometry socket
    """
    return any([s.bl_idname == "NodeSocketGeometry" for s in sockets])


def node_has_geo_socket(node: bpy.types.GeometryNode) -> bool:
    """
    Check if a node has any geometry sockets in its inputs or outputs

    Parameters
    ----------
    node : bpy.types.GeometryNode
        The node to check

    Returns
    -------
    bool
        True if the node has at least one geometry socket
    """
    return any([contains_geo_socket(x) for x in [node.inputs, node.outputs]])


def build_dependency_graph(tree: bpy.types.NodeTree) -> tuple[dict, Counter]:
    """
    Build a graph representing node dependencies and count input connections

    Parameters
    ----------
    tree : bpy.types.NodeTree
        The node tree to analyze

    Returns
    -------
    tuple
        Contains:
            dict
                Mapping of nodes to their dependent nodes
            Counter
                Count of connections for each socket
    """
    dependency_graph = {node: set() for node in tree.nodes}
    socket_input_connection_count = Counter()

    # populate the graph based on node connections
    for link in tree.links:
        dependency_graph[link.from_node].add(link.to_node)
        socket_input_connection_count[link.to_socket] += 1

    return dependency_graph, socket_input_connection_count


def topological_sort(dependency_graph: dict) -> list:
    """
    Sort nodes by their dependencies using a topological sort algorithm

    Parameters
    ----------
    dependency_graph : dict
        Mapping of nodes to their dependent nodes

    Returns
    -------
    list
        Nodes sorted in topological order
    """
    # count incoming connections for each node
    incoming_connection_count = {node: 0 for node in dependency_graph}
    for source_node in dependency_graph:
        for target_node in dependency_graph[source_node]:
            incoming_connection_count[target_node] += 1

    # start with nodes that have no dependencies
    processing_queue = deque(
        [
            node
            for node in incoming_connection_count
            if incoming_connection_count[node] == 0
        ]
    )
    sorted_node_order = []

    # process nodes in breadth-first order
    while processing_queue:
        current_node = processing_queue.popleft()
        sorted_node_order.append(current_node)

        # update counts for nodes that depend on the current node
        for dependent_node in dependency_graph[current_node]:
            incoming_connection_count[dependent_node] -= 1
            # if all dependencies are processed, add to queue
            if incoming_connection_count[dependent_node] == 0:
                processing_queue.append(dependent_node)

    return sorted_node_order


def organize_into_columns(nodes_in_order: list, dependency_graph: dict) -> list:
    """
    Organize nodes into columns based on their dependencies

    Parameters
    ----------
    nodes_in_order : list
        Nodes sorted in topological order
    dependency_graph : dict
        Mapping of nodes to their dependent nodes

    Returns
    -------
    list
        Columns of nodes, where each column is a list of nodes
    """
    node_columns = []
    node_column_assignment = {}

    for node in reversed(nodes_in_order):
        # node goes in column after its furthest dependent
        node_column_assignment[node] = (
            max(
                [
                    node_column_assignment[dependent_node]
                    for dependent_node in dependency_graph[node]
                ],
                default=-1,
            )
            + 1
        )

        # add node to its assigned column
        if node_column_assignment[node] == len(node_columns):
            node_columns.append([node])
        else:
            node_columns[node_column_assignment[node]].append(node)

    # reverse columns to get left-to-right flow
    return list(reversed(node_columns))


def calculate_node_dimensions(
    node: bpy.types.Node, socket_input_connection_count: Counter, interface_scale: float
) -> tuple[float, float]:
    """
    Calculate the visual dimensions of a node

    Parameters
    ----------
    node : bpy.types.Node
        The node to calculate dimensions for
    socket_input_connection_count : Counter
        Counter of connections for each socket
    interface_scale : float
        UI scale factor

    Returns
    -------
    tuple[float, float]
        Width and height of the node
    """
    # height constants for different node elements
    node_header_height = 20
    node_socket_height = 28
    node_property_row_height = 28
    node_vector_input_height = 84

    # count enabled inputs and outputs
    enabled_input_count = len(
        list(filter(lambda input_socket: input_socket.enabled, node.inputs))
    )
    enabled_output_count = len(
        list(filter(lambda output_socket: output_socket.enabled, node.outputs))
    )

    # get properties specific to this node type (not inherited)
    inherited_property_ids = [
        property.identifier
        for base_class in type(node).__bases__
        for property in base_class.bl_rna.properties
    ]

    node_specific_property_count = len(
        [
            property
            for property in node.bl_rna.properties
            if property.identifier not in inherited_property_ids
        ]
    )

    # count vector inputs that need UI widgets (not connected)
    unconnected_vector_input_count = len(
        list(
            filter(
                lambda input_socket: input_socket.enabled
                and input_socket.type == "VECTOR"
                and socket_input_connection_count[input_socket] == 0,
                node.inputs,
            )
        )
    )

    # calculate total node height based on components
    total_node_height = (
        node_header_height
        + (enabled_output_count * node_socket_height)
        + (node_specific_property_count * node_property_row_height)
        + (enabled_input_count * node_socket_height)
        + (unconnected_vector_input_count * node_vector_input_height)
    ) * interface_scale

    return node.width, total_node_height


def position_nodes_in_columns(
    node_columns: list,
    socket_input_connection_count: Counter,
    spacing: typing.Tuple[float, float] = (50, 25),
) -> None:
    """Position nodes in columns with appropriate spacing

    Parameters
    ----------
    node_columns : list
        List of columns, where each column is a list of nodes
    socket_input_connection_count : Counter
        Counter of connections for each socket
    spacing : tuple of float, optional
        Tuple of (horizontal, vertical) spacing between nodes, by default (50, 25)
    """
    interface_scale = bpy.context.preferences.view.ui_scale
    non_geo_offset = 20 + 28 * 2  # header + 2 socket heights

    # position nodes column by column
    position_x = 0
    for column in node_columns:
        widest_node_in_column = 0
        position_y = 0

        for node in column:
            node.update()

            width, height = calculate_node_dimensions(
                node, socket_input_connection_count, interface_scale
            )

            # track widest node for column spacing
            if width > widest_node_in_column:
                widest_node_in_column = width

            # position node
            node.location = (position_x, position_y)

            # adjust position for non-geometry nodes
            if not node_has_geo_socket(node):
                node.location -= Vector((0, non_geo_offset))

            # move down for next node with spacing
            position_y -= height + spacing[1]

        # move right for next column with spacing
        position_x += widest_node_in_column + spacing[0]


def position_special_nodes(
    tree: bpy.types.NodeTree, vertical_offset: float = 100
) -> None:
    """Position special nodes like Group Input and Group Output at the top

    Parameters
    ----------
    tree : bpy.types.NodeTree
        The node tree to modify
    vertical_offset : float, optional
        Vertical offset above the highest node, by default 100

    Returns
    -------
    None
    """
    highest_y_position = max([node.location[1] for node in tree.nodes])
    for special_node_name in ["Group Input", "Group Output"]:
        if special_node_name in tree.nodes:
            special_node = tree.nodes[special_node_name]
            special_node.location = (
                special_node.location[0],
                highest_y_position + vertical_offset,
            )


def cleanup_orphaned_nodes(
    tree: bpy.types.NodeTree, max_iter: int = 100, add_group_input: bool = True
) -> None:
    """Remove nodes that are not connected to anything"""
    to_remove = []
    for _ in range(max_iter):
        for node in tree.nodes:
            if len(node.outputs) == 0:
                continue
            if not any([s.is_linked for s in node.outputs]):
                to_remove.append(node)

        if len(to_remove) == 0:
            break

        for node in to_remove:
            tree.nodes.remove(node)

        to_remove = []

    if add_group_input and "Group Input" not in tree.nodes:
        n_input = tree.nodes.new("NodeGroupInput")
        if "Join Geometry" in tree.nodes:
            tree.links.new(n_input.outputs[0], tree.nodes["Join Geometry"].inputs[0])


def arrange_tree(
    tree: bpy.types.NodeTree,
    spacing: typing.Tuple[float, float] = (50, 25),
    add_group_input: bool = True,
) -> None:
    """Arrange nodes in a node tree based on their dependencies

    Parameters
    ----------
    tree : bpy.types.GeometryNodeTree
        The node tree to arrange
    spacing : tuple of float
        Tuple of (horizontal, vertical) spacing between nodes

    Returns
    -------
    None
        This function modifies the node tree in place

    Notes
    -----
    This function organizes nodes into columns and positions them with appropriate spacing.
    Nodes are arranged from left to right based on their dependencies, with special
    handling for geometry nodes and group input/output nodes.
    """
    cleanup_orphaned_nodes(tree, add_group_input=add_group_input)
    dependency_graph, socket_input_connection_count = build_dependency_graph(tree)
    nodes_in_dependency_order = topological_sort(dependency_graph)
    node_columns = organize_into_columns(nodes_in_dependency_order, dependency_graph)
    position_nodes_in_columns(node_columns, socket_input_connection_count, spacing)
    position_special_nodes(tree)



from .density import Density
from .ensemble import CellPack, Ensemble, StarFile
from .molecule.base import Molecule, MoleculeSelector
from .trajectory import OXDNA, StreamingTrajectory, Trajectory

__all__ = [
    "CellPack",
    "Ensemble",
    "StarFile",
    "Density",
    "Molecule",
    "MoleculeSelector",
    "OXDNA",
    "Trajectory",
    "StreamingTrajectory",
]



import pytest
import molecularnodes as mn
from molecularnodes.nodes import nodes as nodes
from .constants import attributes, data_dir
from .utils import NumpySnapshotExtension

formats = ["mol", "sdf"]


@pytest.mark.parametrize("format", formats)
def test_open(format):
    molecule = mn.Molecule.load(data_dir / f"caffeine.{format}")

    assert molecule.array


@pytest.mark.parametrize("format", formats)
@pytest.mark.parametrize(
    "style",
    [
        mn.StyleBallAndStick(sphere_geometry="Mesh"),
        mn.StyleSpheres(geometry="Mesh"),
        mn.StyleSurface(),
    ],
)
def test_load(snapshot_custom: NumpySnapshotExtension, format, style):
    mol = mn.Molecule.load(data_dir / f"caffeine.{format}").add_style(style=style)
    assert mol.object

    for attribute in attributes:
        try:
            print(f"{attribute=}")
            assert snapshot_custom == mol.named_attribute(attribute, evaluate=True)
        except AttributeError as e:
            assert e
